#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

class Token(object):
    def __init__(self, token_type, lexeme, literal, line):
        self.token_type = token_type
        self.lexeme     = lexeme
        self.literal    = literal
        self.line       = line

    def __str__(self):
        return 'Type:"{self.token_type}" Lexeme:"{self.lexeme}" Lit:"{self.literal}"'.format(self=self)



#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

import docopt
import fileinput

from .interpreter import Interpreter

__version__ = '0.0.2'

__doc__ = """sassy - interpreter for the sassy macro language

Usage:
    sassy [-i | --ignore] [<file>...]
    sassy (-h | --help)
    sassy (-v | --version)

Options:
    -i, --ignore        Ignore whitespace outside of macro tags
    -h, --help          Show help description.
    -v, --version       Show version.
"""


def main():
    # needs to be outside of try block to let docopt handle bad input
    arguments = docopt.docopt(__doc__, version='sassy v' + __version__)

    try:
        Interpreter(arguments).process_input(fileinput)
    except FileNotFoundError:
        print('File not found!')


if __name__ == '__main__':
    main()



#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class TokenType(Enum):
    # sassy - specific keywords
    MACRO_START     = 1
    MACRO_END       = 2
    LET             = 3
    LOOP_START      = 4
    PROC_END        = 5
    EXEC            = 6

    # single character tokens
    LEFT_PAREN      = 7
    RIGHT_PAREN     = 8
    LEFT_BRACE      = 9
    RIGHT_BRACE     = 10
    COMMA           = 11
    DOT             = 12
    MINUS           = 13
    PLUS            = 14
    SEMICOLON       = 15
    SLASH           = 16
    STAR            = 17
    PERCENT         = 18
    AMPERSAND       = 19
    DOLLAR          = 20
    HASH            = 21
    AT              = 22
    PIPE            = 23
    QUESTION_MARK   = 24
    COLON           = 25
    LEFT_SQR_PAREN  = 26
    RIGHT_SQR_PAREN = 27

    SPACE           = 28
    TAB             = 29
    CARRIAGE_RETURN = 30
    NEW_LINE        = 31

    # one or two character tokens
    BANG            = 32
    BANG_EQUAL      = 33
    EQUAL           = 34
    EQUAL_EQUAL     = 35
    GREATER         = 36
    GREATER_EQUAL   = 37
    LESS            = 38
    LESS_EQUAL      = 39

    # literals
    IDENTIFIER      = 40
    STRING          = 41
    NUMBER          = 42

    # other keywords
    AND             = 43
    CLASS           = 44
    ELSE            = 45
    FALSE           = 46
    FOR             = 47
    IF              = 48
    NIL             = 49
    OR              = 50
    PRINT           = 51
    RETURN          = 52
    SUPER           = 53
    TRUE            = 54
    VAR             = 55
    WHILE           = 56
    EOF             = 57



#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

from .error import *
from .token import Token
from .tokentype import TokenType


class Scanner(object):

    def __init__(self, source):
        self.source = source
        self.tokens = []

        self._start = 0
        self._current = 0
        self._line = 1
        self.keywords = {
            '%macro': TokenType.MACRO_START,
            '%mend': TokenType.MACRO_END,
            '%let': TokenType.LET,
            '%procloop': TokenType.LOOP_START,
            '%pend': TokenType.PROC_END,
            '%exec': TokenType.EXEC
        }

    def scan(self):
        while not self._is_at_end():
            self._start = self._current
            self.scan_token()

        return self.tokens

    def scan_token(self):
        c = self._advance()
        if c == '(':
            self.add_token(TokenType.LEFT_PAREN)
        elif c == ')':
            self.add_token(TokenType.RIGHT_PAREN)
        elif c == '{':
            self.add_token(TokenType.LEFT_BRACE)
        elif c == '}':
            self.add_token(TokenType.RIGHT_BRACE)
        elif c == ',':
            self.add_token(TokenType.COMMA)
        elif c == '.':
            self.add_token(TokenType.DOT)
        elif c == '-':
            self.add_token(TokenType.MINUS)
        elif c == '+':
            self.add_token(TokenType.PLUS)
        elif c == ';':
            self.add_token(TokenType.SEMICOLON)
        elif c == '*':
            self.add_token(TokenType.STAR)
        # elif c == '%':
        #     self.add_token(TokenType.PERCENT)
        elif c == '$':
            self.add_token(TokenType.DOLLAR)
        elif c == '#':
            self.add_token(TokenType.HASH)
        elif c == '@':
            self.add_token(TokenType.AT)
        elif c == '|':
            self.add_token(TokenType.PIPE)
        elif c == '&':
            self.add_token(TokenType.AMPERSAND)
        elif c == ':':
            self.add_token(TokenType.COLON)
        elif c == '?':
            self.add_token(TokenType.QUESTION_MARK)
        elif c == '[':
            self.add_token(TokenType.LEFT_SQR_PAREN)
        elif c == ']':
            self.add_token(TokenType.RIGHT_SQR_PAREN)
        elif c == '!':
            self.add_token(
                TokenType.BANG_EQUAL if self._match('=') else TokenType.BANG)
        elif c == '=':
            self.add_token(
                TokenType.EQUAL_EQUAL if self._match('=') else TokenType.EQUAL)
        elif c == '<':
            self.add_token(
                TokenType.LESS_EQUAL if self._match('=') else TokenType.LESS)
        elif c == '>':
            self.add_token(
                TokenType.GREATER_EQUAL if self._match('=') else TokenType.GREATER)
        elif c == '/':
            if self._match('/'):
                while (self._peek() != '\n' and not self._is_at_end()):
                    self._advance()
            else:
                self.add_token(TokenType.SLASH)
        elif c == ' ':
            self.add_token(TokenType.SPACE)
        elif c == '\r':
            self.add_token(TokenType.CARRIAGE_RETURN)
        elif c == '\t':
            self.add_token(TokenType.TAB)
        elif c == '\n':
            self.add_token(TokenType.NEW_LINE)
            self._line += 1
        elif c == '"':
            self.identifier()
        elif c == '\'':
            self.identifier()
        elif self._is_digit(c):
            self.number()
        elif self._is_alpha(c):
            self.identifier()
        else:
            self._scan_error(self._line, self._current, 'Unexpected character.')

    def _scan_error(self, line, where, message):
        errorText = '[line {line} char {where}] Error: {message}'.format(
            line=line,
            where=where,
            message=message
        )
        errorSource  = self.source
        errorPointer = ' ' * (int(where) - 1)
        errorPointer += '^'

        errorMsg = errorText + '\n' + errorSource + errorPointer + '\n'
        raise ScanError(errorMsg)

    def _is_at_end(self):
        return self._current >= len(self.source)

    def identifier(self):
        while self._is_alpha_numeric(self._peek()):
            self._advance()
        text = self.source[self._start:self._current]
        token_type = self.keywords.get(text)
        if token_type is None:
            token_type = TokenType.IDENTIFIER

        self.add_token(token_type)

    def number(self):
        while self._is_digit(self._peek()):
            self._advance()

        if self._peek() == '.' and self._is_digit(self._peek_next()):
            # consume the dot
            self._advance()
            while self._is_digit(self._peek()):
                self._advance()

        self.add_token(
            TokenType.NUMBER, float(self.source[self._start:self._current]))

    def string(self):
        while self._peek() != '"' and not self._is_at_end():
            if self._peek == '\n':
                self._line += 1
            self._advance()

        # unterminated string
        if self._is_at_end():
            self._scan_error(self._line, self._current, 'Unterminated string.')

        self._advance()
        value = self.source[self._start + 1:self._current - 1]
        self.add_token(TokenType.STRING, value)

    def _is_digit(self, c):
        return c.isdigit()

    def _is_alpha(self, c):
        return c.isalpha() or c == '_' or c == '%'

    def _is_alpha_numeric(self, c):
        return self._is_alpha(c) or self._is_digit(c)

    def _match(self, expected):
        if self._is_at_end():
            return False

        if self.source[self._current] != expected:
            return False

        self._current += 1
        return True

    def _peek(self):
        if self._current >= len(self.source):
            return '\0'
        return self.source[self._current]

    def _peek_next(self):
        if (self._current + 1) >= len(self.source):
            return '\0'
        return self.source[self._current + 1]

    def _advance(self):
        self._current += 1
        return self.source[self._current - 1]

    def add_token(self, token_type, literal=None):
        text = self.source[self._start:self._current]
        self.tokens.append(Token(token_type, text, literal, self._line))



#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

import sys

from .scanner import Scanner
from .parser import Parser
from .translator import Translator
from .error import *
from .token import Token
from .tokentype import TokenType


class Interpreter(object):

    def __init__(self, arguments):
        # program options
        self._fileList = arguments['<file>']
        self._ignoreWhitespace = arguments['--ignore']

        # scanner outputs
        self.tokenList = []
        self.tokenListSize = 0

        # parser outputs
        self.translationSteps = {}

    def process_input(self, source):

        # scan input
        try:
            for line in source.input(self._fileList):
                self._scan_source(line)
        except ScanError as err:
            print("Error[scanner]: Bad input provided", file=sys.stderr)
            print(err.message, file=sys.stderr)
            sys.exit(-1)

        # remove EOF characters resulting from paged input
        self.tokenList = [token for token in self.tokenList if token.token_type != TokenType.EOF]
        self.tokenList.append(Token(TokenType.NEW_LINE, '\n', None, -1))  # add newline to end
        self.tokenListSize = len(self.tokenList)

        # parse tokens
        if bool(self.tokenList):
            try:
                self.translationSteps = Parser(
                    self.tokenList,
                    self.tokenListSize,
                    self._ignoreWhitespace
                ).parse()
            except BadDefinitionError as err:
                print("Error[parser]: BadDefinitionError", file=sys.stderr)
                print(err.message, file=sys.stderr)
                sys.exit(-1)
            except ExistingVariableError as err:
                print("Error[parser]: ExistingVariableError", file=sys.stderr)
                print(err.message, file=sys.stderr)
                sys.exit(-1)
            except ExistingMacroError as err:
                print("Error[parser]: ExistingMacroError", file=sys.stderr)
                print(err.message, file=sys.stderr)
                sys.exit(-1)
            except NonExistingMacroError as err:
                print("Error[parser]: NonExistingMacroError", file=sys.stderr)
                print(err.message, file=sys.stderr)
                sys.exit(-1)
            except LoopParameterError as err:
                print("Error[parser]: LoopParameterError", file=sys.stderr)
                print(err.message, file=sys.stderr)
                sys.exit(-1)

        # translate steps
        if bool(self.translationSteps):
            try:
                Translator(self.translationSteps).translate()
            except IncorrectParametersError as err:
                print('Error[translator]: IncorrectParametersError', file=sys.stderr)
                print(err.message, file=sys.stderr)
                sys.exit(-1)
            except LoopParameterError as err:
                print('Error[translator]: LoopParameterError', file=sys.stderr)
                print(err.message, file=sys.stderr)
                sys.exit(-1)

    def _scan_source(self, source):
        for token in Scanner(source).scan():
            self.tokenList.append(token)



#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class ScanError(Error):
    """Exception raised for errors occuring whilst scanning the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class BadDefinitionError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class ExistingMacroError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class NonExistingMacroError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class IncorrectParametersError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class ExistingVariableError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class LoopParameterError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message



#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

from .token import Token
from .tokentype import TokenType
from .error import *
# import json  # used printing debug data in _consolidateOutputs()


class Parser(object):

    def __init__(self, tokens, numTokens, optIgnoreWhitespace):
        self.tokenList     = tokens
        self.tokenListSize = numTokens - 1

        self.outputSteps = {}  # operations keyed by order number
        self.macroDef    = {}  # macros keyed by macro name
        self.varDef      = {}  # variables keyed by variable name

        self._currentTokenIndex = 0
        self._stepOrderNumber   = 0  # order number for the output steps
        self._maxMacroParams    = 25  # max number of macro parameters

        # parsing options
        self._ignoreWhitespace  = optIgnoreWhitespace

    def parse(self):
        """
        Parse each line of tokens and process the
        corresponding context from the list below:

        1. Plain text: This is any line not including a keyword
        2. Macro: Macro start/end
        3. Let: Variable definition
        4. Exec: Execution of a macro
        5. Procloop: Execution of a loop
        """
        tokenStack = []
        while not self._is_end_of_loop():
            # get current line
            result, tokenStack = self._get_line()

            if len(tokenStack) > 0:
                # determine type of line (contains macro statement, plaintext)
                lineTypeToken = self._get_line_type(tokenStack)
                cleanTokenStack = self._remove_line_whitespace(tokenStack)

                # handle block
                if not self._match(lineTypeToken, self._get_keyword_token_list()):
                    self._processPlainText(cleanTokenStack)

                elif self._match(lineTypeToken, [TokenType.MACRO_START]):
                    self._processMacro(cleanTokenStack)

                elif self._match(lineTypeToken, [TokenType.LET]):
                    self._processLet(cleanTokenStack)

                elif self._match(lineTypeToken, [TokenType.LOOP_START]):
                    self._processLoop(cleanTokenStack)

                elif self._match(lineTypeToken, [TokenType.EXEC]):
                    self._processExec(cleanTokenStack)
                else:
                    errorMsg = 'Unknown char in line "' + ' '.join(tokenStack) + '"'
                    raise BadDefinitionError(errorMsg)

        return self._consolidateOutputs()

    def _advance(self):
        """
        Returns token at position _currentTokenIndex
        Returns -1 if at the end of the tokenList.
        """
        if self._is_end_of_loop():
            return -1
        else:
            self._currentTokenIndex += 1
            return self.tokenList[self._currentTokenIndex]

    def _consolidateOutputs(self):
        output = {
            'outputSteps': self.outputSteps,
            'macroDef': self.macroDef,
            'varDef': self.varDef
        }
        # uncomment below to print final parse tree
        # print('\n\n_consolidateOutputs():')
        # print(json.dumps(output, indent=3))
        return output

    def _get_line(self):
        """
        Returns the next line of tokens as a list,
        and a result integer (0 if successful, -1 on EOF).
        """
        tokenStack = []
        token = self._get_current_token()

        # check if EOF
        if self._is_end_of_loop():
            return (-1, tokenStack)

        while not self._match(token, [TokenType.NEW_LINE]):
            tokenStack.append(token)
            token = self._advance()
            if token == -1:
                return (-1, tokenStack)
        tokenStack.append(token)  # add new line char
        self._advance()  # move past newline

        return (0, tokenStack)

    def _get_line_type(self, tokenStack):
        """
        Returns the first keyword token
        in the list if it exists, the first
        token in the list otherwise if plaintext.
        Returns -1 , if the tokenStack is empty
        """
        if len(tokenStack):
            for token in tokenStack:
                if self._match(token, self._get_keyword_token_list()):
                    return token
            return tokenStack[0]
        else:
            return -1

    def _remove_line_whitespace(self, tokenStack):
        """
        Returns a whitespace-free list of tokens,
        if line contains a keyword.
        Returns unchanged list of tokens if the line
        contains just plaintext.
        """
        isKeywordLine = 0
        keywordToken = ''
        outputTokenStack = []

        if len(tokenStack):
            # determine if line contains a keyword
            token = self._get_line_type(tokenStack)
            if self._match(token, self._get_keyword_token_list()):
                isKeywordLine = 1
                keywordToken = token

            if isKeywordLine == 1:
                # let tags need to include all whitespace between "=" and ";"
                if self._match(keywordToken, [TokenType.LET]):
                    isPastEqualSymbol = 0
                    for token in tokenStack:
                        if not self._match(token, [TokenType.EQUAL]):
                            if self._match(token, [TokenType.SEMICOLON]):
                                outputTokenStack.append(token)
                                break
                            else:
                                if isPastEqualSymbol == 1:
                                    outputTokenStack.append(token)
                                else:
                                    if not self._match(token, self._get_whitespace_token_list()):
                                        outputTokenStack.append(token)
                        else:
                            isPastEqualSymbol = 1
                            outputTokenStack.append(token)
                            continue
                    return outputTokenStack
                else:
                    # strip all whitespace otherwise
                    for token in tokenStack:
                        if not self._match(token, self._get_whitespace_token_list()):
                            outputTokenStack.append(token)
                    return outputTokenStack
            else:
                return tokenStack
        else:
            return -1

    def _is_end_of_loop(self):
        """
        Returns True if at the end of the list of source tokens
        and False otherwise.
        """
        return self._currentTokenIndex >= self.tokenListSize

    def _match(self, token, tokenTypeList):
        """
        Returns True if the passed token's type
        is in the list of passed types, and False
        otherwise.
        """
        if token.token_type in tokenTypeList:
            return True
        else:
            return False

    def _increment_step_order(self):
        self._stepOrderNumber += 1

    def _decrement_step_order(self):
        self._stepOrderNumber -= 1

    def _increment_current_index(self):
        self._currentTokenIndex += 1

    def _decrement_current_index(self):
        self._currentTokenIndex -= 1

    def _get_current_token(self):
        return self.tokenList[self._currentTokenIndex]

    def _get_whitespace_token_list(self):
        return [TokenType.SPACE,
                TokenType.TAB,
                TokenType.CARRIAGE_RETURN,
                TokenType.NEW_LINE]

    def _get_keyword_token_list(self):
        return [TokenType.MACRO_START,
                TokenType.MACRO_END,
                TokenType.LET,
                TokenType.LOOP_START,
                TokenType.PROC_END,
                TokenType.EXEC]

    def _processPlainText(self, tokenStack):
        """
        Process plain text context.
        """
        outputStep = {}
        isNotWhitespaceLine = 0

        # concat current token to output string
        plainTextLine = ''
        for token in tokenStack:
            if not self._match(token, self._get_whitespace_token_list()):
                isNotWhitespaceLine = 1
            plainTextLine += token.lexeme

        # if option is set, ignore all plaintext lines
        # that are composed only of whitespace tokens
        if (not isNotWhitespaceLine) and self._ignoreWhitespace:
            return

        # store plain text data
        outputStep["type"] = "plaintext"
        outputStep["text"] = plainTextLine

        self._increment_step_order()
        self.outputSteps[self._stepOrderNumber] = outputStep

    def _processLet(self, tokenStack):
        """
        Process %let statement of the form

        %let varName = varVal ;
        """
        letTokenIndex         = 0
        varNameTokenIndex     = 1
        equalsTokenIndex      = 2
        varValStartTokenIndex = 3
        semiColonTokenIndex   = -1

        varName = ''
        varVal  = ''

        if not self._match(tokenStack[varNameTokenIndex], [TokenType.IDENTIFIER]):
            errorMsg = 'Bad let tag definition. Invalid variable name.'
            raise BadDefinitionError(errorMsg)
        varName = tokenStack[varNameTokenIndex].lexeme
        if varName in self.varDef:
            errorMsg = 'Variable "' + varName + '" already exists.'
            raise ExistingVariableError(errorMsg)

        # check for correct let tag ending
        if not self._match(tokenStack[semiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" in let tag "' + varName + '".'
            raise BadDefinitionError(errorMsg)

        # check let tag only has one '=', and process subsequent tokens
        filterList = [
            TokenType.EQUAL
        ]
        checkList = [t.lexeme for t in tokenStack if self._match(t, filterList)]
        if len(checkList) > 1:
            errorMsg = 'Bad let tag definition. More than one "=" found.'
            raise BadDefinitionError(errorMsg)

        # get variable value
        varVal = ''
        varValList = [t for t in tokenStack[varValStartTokenIndex:]]
        for v in varValList:
            if not self._match(v, [TokenType.SEMICOLON]):
                varVal += v.lexeme

        # add variable definition to dicionary
        self.varDef[varName] = varVal

    def _processMacro(self, tokenStack):
        """
        Process macro context of the form:

        %macro <macroname> (<param1>, <param2>, ...);
            <macro block>
        %mend;
        """
        macroStartTokenIndex      = 0
        macroNameTokenIndex       = 1
        macroOpenParenTokenIndex  = 2
        macroCloseParenTokenIndex = -2
        macroSemiColonTokenIndex  = -1

        macroStep   = {}
        macroName   = ''
        macroParams = []
        macroBody   = ''

        # extract macro name
        if not self._match(tokenStack[macroNameTokenIndex], [TokenType.IDENTIFIER]):
            errorMsg = 'Bad macro definition'
            raise BadDefinitionError(errorMsg)
        macroName = tokenStack[macroNameTokenIndex].lexeme

        # check for correct macro signature ending
        if not self._match(tokenStack[macroSemiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" in opening tag of macro "' + macroName + '"'
            raise BadDefinitionError(errorMsg)

        # extract macro parameters
        filterParamsList = [
            TokenType.LEFT_PAREN,
            TokenType.RIGHT_PAREN,
            TokenType.COMMA,
            TokenType.SEMICOLON,
            TokenType.NEW_LINE
        ]
        paramList = [p for p in tokenStack[macroOpenParenTokenIndex:]
                        if not self._match(p, filterParamsList)]

        # check parameters for validity
        acceptedParamsList = [
            TokenType.IDENTIFIER
        ]
        for p in paramList:
            if not self._match(p, acceptedParamsList):
                errorMsg = 'Incorrect parameter in macro "' + macroName + '"'
                raise BadDefinitionError(errorMsg)
            else:
                macroParams.append(p.lexeme)

        # macro body
        result, tokenStack = self._get_line()
        lineTypeToken = self._get_line_type(tokenStack)
        tokenStack = self._remove_line_whitespace(tokenStack)

        if result != -1:
            while not self._match(lineTypeToken, [TokenType.MACRO_END]):
                for token in tokenStack:
                    macroBody += token.lexeme

                result, tokenStack = self._get_line()
                lineTypeToken = self._get_line_type(tokenStack)
                tokenStack = self._remove_line_whitespace(tokenStack)

                if result == -1 and not len(tokenStack):
                    errorMsg = 'No closing tag for macro "' + macroName + '"'
                    raise BadDefinitionError(errorMsg)
                if result == -1 and len(tokenStack):
                    break
        else:
            errorMsg = 'No closing tag for macro "' + macroName + '"'
            raise BadDefinitionError(errorMsg)

        # ensure correctness of remaining macro closing tag '%mend;'
        if len(tokenStack) != 2:
            errorMsg = 'Bad closing tag for macro "' + macroName + '"'
            raise BadDefinitionError(errorMsg)
        if not self._match(tokenStack[macroSemiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" on closing tag for macro "' + macroName + '"'
            raise BadDefinitionError(errorMsg)

        # store in macroDef{} and in outputSteps{}
        macroStep['type']       = 'macro'
        macroStep['parameters'] = macroParams
        macroStep['text']       = macroBody

        if macroName in self.macroDef:
            errorMsg = 'Macro "' + macroName + '" already exists'
            raise ExistingMacroError(errorMsg)

        # add macro step
        self.macroDef[macroName] = macroStep

    def _processNestedLoop(self, tokenStack):
        procloopOpenParenTokenIndex  = 1
        procloopCloseParenTokenIndex = -3
        procloopCounterVarTokenIndex = -2
        procloopSemiColonTokenIndex  = -1

        procloopMaxParameters = 3

        loopStart   = 0
        loopEnd     = 0
        loopInc     = 0
        loopVar     = ''
        loopSteps   = {}  # holds the current loop steps
        outputSteps = {}  # gets returned as the top-level output step

        # check for correct procloop signature ending
        if not self._match(tokenStack[procloopSemiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" in procloop signature.'
            raise BadDefinitionError(errorMsg)

        if not self._match(tokenStack[procloopCounterVarTokenIndex], [TokenType.IDENTIFIER]):
            errorMsg = 'Missing counter variable in procloop.'
            raise BadDefinitionError(errorMsg)
        else:
            # get loop counter variable name
            loopVar = tokenStack.pop(procloopCounterVarTokenIndex).lexeme

        # extract procloop parameter
        paramList = self._extract_params(tokenStack)

        if len(paramList) > procloopMaxParameters or len(paramList) == 0:
            errorMsg = 'Incorrect number of procloop tag parameters passed.'
            raise BadDefinitionError(errorMsg)

        # set loop params
        if len(paramList) == 1:
            loopStart = paramList[0].lexeme
        if len(paramList) == 2:
            loopStart = paramList[0].lexeme
            loopEnd   = paramList[1].lexeme
        if len(paramList) == 3:
            loopStart = paramList[0].lexeme
            loopEnd   = paramList[1].lexeme
            loopInc   = paramList[2].lexeme

        # procloop body
        loopStepsCounter = 0

        result, tokenStack = self._get_line()
        lineTypeToken = self._get_line_type(tokenStack)
        tokenStack = self._remove_line_whitespace(tokenStack)
        if result != -1:
            while not self._match(lineTypeToken, [TokenType.PROC_END]):

                if not self._match(lineTypeToken, [TokenType.LOOP_START]):
                    # plain text
                    loopStepsCounter += 1
                    loopSteps[loopStepsCounter] = self._processLoopPlainText(tokenStack)
                else:
                    loopStepsCounter += 1
                    loopSteps[loopStepsCounter] = self._processNestedLoop(tokenStack)

                result, tokenStack = self._get_line()
                lineTypeToken = self._get_line_type(tokenStack)
                tokenStack = self._remove_line_whitespace(tokenStack)
                if result == -1 and not len(tokenStack):
                    errorMsg = 'No closing tag for procloop'
                    raise BadDefinitionError(errorMsg)
                if result == -1 and len(tokenStack):
                    break
        else:
            errorMsg = 'No closing tag for procloop'
            raise BadDefinitionError(errorMsg)

        # ensure correctness of procloop closing tag '%pend;'
        if len(tokenStack) != 2:
            errorMsg = 'Bad closing tag for procloop'
            raise BadDefinitionError(errorMsg)
        if not self._match(tokenStack[procloopSemiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" on closing tag for procloop'
            raise BadDefinitionError(errorMsg)

        # build outputs
        outputSteps['type']      = 'loop'
        outputSteps['loopStart'] = loopStart
        outputSteps['loopEnd']   = loopEnd
        outputSteps['loopInc']   = loopInc
        outputSteps['loopVar']   = loopVar
        outputSteps['loopSteps'] = loopSteps

        return outputSteps

    def _processLoopPlainText(self, tokenStack):
        loopPlaintextDict = {}
        loopPlaintextBuffer = ''

        for token in tokenStack:
            loopPlaintextBuffer += token.lexeme

        # build outputs
        loopPlaintextDict['type'] = 'plaintext'
        loopPlaintextDict['text'] = loopPlaintextBuffer

        return loopPlaintextDict

    def _extract_params(self, tokenStack):
        """
        Returns list of tokens that are either
        of TokenType NUMBER or IDENTIFIER
        """
        filterParamsList = self._get_keyword_token_list()
        filterParamsList.append(TokenType.LEFT_PAREN)
        filterParamsList.append(TokenType.SEMICOLON)
        filteredList = [p for p in tokenStack if not self._match(p, filterParamsList)]

        paramAcc  = ''
        paramList = []
        for token in filteredList:
            if not self._match(token, [TokenType.COMMA, TokenType.RIGHT_PAREN]):
                paramAcc += token.lexeme
            else:
                # avoid adding empty string
                if paramAcc != '':
                    paramList.append(Token(TokenType.IDENTIFIER, paramAcc, None, -1))
                    paramAcc = ''  # reset param accumulator
                if self._match(token, [TokenType.RIGHT_PAREN]):
                    break

        return paramList

    def _processLoop(self, tokenStack):
        """
        Process loop context of the form:

        %procloop (<number of iterations>) <counters variable>;
            <proc loop body>
        %pend;

        The <proc loop body> may contain nested procloops.

        The <counter variable> is an identifier
        used to reference the count of each iteration
        of the loop when executing.
        """
        procloopOpenParenTokenIndex  = 1
        procloopCloseParenTokenIndex = -3
        procloopCounterVarTokenIndex = -2
        procloopSemiColonTokenIndex  = -1

        procloopMaxParameters = 3

        loopStart   = 0
        loopEnd     = 0
        loopInc     = 0
        loopVar     = ''
        loopSteps   = {}  # holds the current loop steps
        outputSteps = {}  # gets returned as the top-level output step

        # check for correct procloop signature ending
        if not self._match(tokenStack[procloopSemiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" in procloop signature.'
            raise BadDefinitionError(errorMsg)

        if not self._match(tokenStack[procloopCounterVarTokenIndex], [TokenType.IDENTIFIER]):
            errorMsg = 'Missing counter variable in procloop.'
            raise BadDefinitionError(errorMsg)
        else:
            # get loop counter variable name
            loopVar = tokenStack.pop(procloopCounterVarTokenIndex).lexeme

        # extract procloop parameter
        paramList = self._extract_params(tokenStack)

        # check number of parameters
        if len(paramList) > procloopMaxParameters or len(paramList) == 0:
            errorMsg = 'Incorrect number of procloop tag parameters passed.'
            raise BadDefinitionError(errorMsg)

        # set loop params
        if len(paramList) == 1:
            loopStart = paramList[0].lexeme
        if len(paramList) == 2:
            loopStart = paramList[0].lexeme
            loopEnd   = paramList[1].lexeme
        if len(paramList) == 3:
            loopStart = paramList[0].lexeme
            loopEnd   = paramList[1].lexeme
            loopInc   = paramList[2].lexeme

        # procloop body
        loopStepsCounter = 0

        result, tokenStack = self._get_line()
        lineTypeToken = self._get_line_type(tokenStack)
        tokenStack = self._remove_line_whitespace(tokenStack)

        if result != -1:
            while not self._match(lineTypeToken, [TokenType.PROC_END]):

                if not self._match(lineTypeToken, [TokenType.LOOP_START]):
                    # plain text
                    loopStepsCounter += 1
                    loopSteps[loopStepsCounter] = self._processLoopPlainText(tokenStack)
                else:
                    loopStepsCounter += 1
                    loopSteps[loopStepsCounter] = self._processNestedLoop(tokenStack)

                result, tokenStack = self._get_line()
                lineTypeToken = self._get_line_type(tokenStack)
                tokenStack = self._remove_line_whitespace(tokenStack)

                if result == -1 and not len(tokenStack):
                    errorMsg = 'No closing tag for procloop'
                    raise BadDefinitionError(errorMsg)
                if result == -1 and len(tokenStack):
                    break
        else:
            errorMsg = 'No closing tag for procloop'
            raise BadDefinitionError(errorMsg)

        # ensure correctness of procloop closing tag '%pend;'
        if len(tokenStack) != 2:
            errorMsg = 'Bad closing tag for procloop'
            raise BadDefinitionError(errorMsg)
        if not self._match(tokenStack[procloopSemiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" on closing tag for procloop'
            raise BadDefinitionError(errorMsg)

        # store loop in dict
        outputSteps['type']      = 'loop'
        outputSteps['loopStart'] = loopStart
        outputSteps['loopEnd']   = loopEnd
        outputSteps['loopInc']   = loopInc
        outputSteps['loopVar']   = loopVar
        outputSteps['loopSteps'] = loopSteps

        self._increment_step_order()
        self.outputSteps[self._stepOrderNumber] = outputSteps

    def _processExec(self, tokenStack):
        """
        Process exec statement of the form:

        %exec <macroname>(<param1>, <param2>, ...);
        """
        execMacroNameTokenIndex  = 1
        execOpenParenTokenIndex  = 2
        execCloseParenTokenIndex = -2
        execSemiColonTokenIndex  = -1

        execStep      = {}
        execMacroName = ''
        execParams    = []

        # extract exec macro name
        if not self._match(tokenStack[execMacroNameTokenIndex], [TokenType.IDENTIFIER]):
            errorMsg = 'Bad exec tag definition'
            raise BadDefinitionError(errorMsg)
        execMacroName = tokenStack[execMacroNameTokenIndex].lexeme

        # check for correct exec signature ending
        if not self._match(tokenStack[execSemiColonTokenIndex], [TokenType.SEMICOLON]):
            errorMsg = 'Missing ";" in exec tag "' + execMacroName + '"'
            raise BadDefinitionError(errorMsg)

        execParams = [p.lexeme for p in self._extract_params(tokenStack[execOpenParenTokenIndex:])]

        # check target macro for exec exists
        if execMacroName not in self.macroDef:
            errorMsg = 'Exec target macro "' + execMacroName + '" does not exist'
            raise NonExistingMacroError(errorMsg)

        # check exec params match required macro params
        if len(self.macroDef[execMacroName]['parameters']) != len(execParams):
            errorMsg = 'Exec tag params do not match target macro params'
            raise BadDefinitionError(errorMsg)

        # add exec to dicionary
        self._increment_step_order()
        execStep['type']       = 'exec'
        execStep['macroName']  = execMacroName
        execStep['parameters'] = execParams
        self.outputSteps[self._stepOrderNumber] = execStep

    def _debug_print(self, token):
        print('* {self._currentTokenIndex}: {token.token_type} \
            {token.lexeme}'.format(self=self, token=token))



#Copyright 2018 Jonathan Sacramento

#This file is part of sassy.
#
#sassy is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#sassy is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with sassy. If not, see <http://www.gnu.org/licenses/>.

import sys

from .error import *


class Translator(object):

    def __init__(self, translationSteps):
        self.translationSteps = translationSteps
        self.totalSteps       = len(translationSteps['outputSteps'])
        self._currentStep     = 0
        self._loopStack       = {}  # used to keep track of nested loop counts

    def translate(self):
            self.translate_steps()

    def translate_steps(self):
        """
        Iterate through, and process each output step.

        This function will process each output step
        in the output dictionary from the parser stage.
        Each step will result in text that is printed
        to standard output.
        """

        # replace variable names first
        for var in self.translationSteps['varDef']:
            varVal    = self.translationSteps['varDef'][var]
            newVarVal = self._replace_params(self.translationSteps['varDef'], varVal)
            self.translationSteps['varDef'][var] = newVarVal

        self._currentStep = 1  # steps start at one
        while self._currentStep <= self.totalSteps:

            if self.translationSteps['outputSteps'][self._currentStep]['type'] == 'plaintext':
                self._translate_plaintext(self._currentStep)

            if self.translationSteps['outputSteps'][self._currentStep]['type'] == 'exec':
                self._translate_exec(self._currentStep)

            if self.translationSteps['outputSteps'][self._currentStep]['type'] == 'loop':
                self._translate_loop(
                    loopStart=self.translationSteps['outputSteps'][self._currentStep]['loopStart'],
                    loopEnd=self.translationSteps['outputSteps'][self._currentStep]['loopEnd'],
                    loopInc=self.translationSteps['outputSteps'][self._currentStep]['loopInc'],
                    loopVar=self.translationSteps['outputSteps'][self._currentStep]['loopVar'],
                    loopSteps=self.translationSteps['outputSteps'][self._currentStep]['loopSteps']
                )

            self._currentStep += 1

    def _translate_plaintext(self, stepNum):
        print(self.translationSteps['outputSteps'][stepNum]['text'], end='', file=sys.stdout)

    def _translate_exec(self, stepNum):
        """
        Translate exec step
        """
        # we know the macro exists
        macroName = self.translationSteps['outputSteps'][stepNum]['macroName']

        # replace params and vars in macro body
        macroParamNames = self.translationSteps['macroDef'][macroName]['parameters']
        macroParamVals = self.translationSteps['outputSteps'][stepNum]['parameters']

        # check matching number of parameters between macro def and exec
        if len(macroParamNames) != len(macroParamVals):
            errorMsg = str(macroName) + ' expects ' + str(len(macroParamNames)) + ' parameters'
            raise IncorrectParametersError(errorMsg)

        # map params to values (positional)
        paramMap = {}
        for i in range(len(macroParamNames)):
            paramMap[macroParamNames[i]] = macroParamVals[i]

        # replace parameter values in macro body
        macroText = self.translationSteps['macroDef'][macroName]['text']
        macroText = self._replace_names(paramMap, macroText)

        # replace var values in macro body
        macroText = self._replace_names(self.translationSteps['varDef'], macroText)

        # print output macro text
        print(macroText, end='', file=sys.stdout)

    def _translate_loop(self, **kwargs):
        """
        Translate loop step
        """
        loopStart = int(self._convert_param_names(kwargs['loopStart']))
        loopEnd   = int(self._convert_param_names(kwargs['loopEnd']))
        loopInc   = int(self._convert_param_names(kwargs['loopInc']))
        loopVar   = kwargs['loopVar']
        loopSteps = kwargs['loopSteps']

        # correct loop parameters to avoid range() error
        if loopInc == 0:
            loopInc = 1

        if loopEnd == 0:
            loopEnd = loopStart
            loopStart = 0

        # print output macro text
        if loopStart == 0 and loopEnd == 0:
            errorMsg = 'procloop requires integer parameters.'
            raise LoopParameterError(errorMsg)
        else:
            try:
                for counter in range(loopStart, loopEnd, loopInc):
                    # update global loop stack
                    self._loopStack[loopVar] = counter
                    for step in loopSteps.keys():
                        if loopSteps[step]['type'] == 'plaintext':
                            # replace variable values in loop body
                            output = loopSteps[step]['text']
                            output = self._replace_names(self._loopStack, output)
                            output = self._replace_names(self.translationSteps['varDef'], output)
                            print(output, end='', file=sys.stdout)
                        if loopSteps[step]['type'] == 'loop':
                            self._translate_loop(
                                loopStart=loopSteps[step]['loopStart'],
                                loopEnd=loopSteps[step]['loopEnd'],
                                loopInc=loopSteps[step]['loopInc'],
                                loopVar=loopSteps[step]['loopVar'],
                                loopSteps=loopSteps[step]['loopSteps']
                            )
            except:
                errorMsg = 'procloop requires integer parameters'
                raise LoopParameterError(errorMsg)

    def _replace_names(self, nameMap, text):
        for name in nameMap:
            formatName = "&" + name + "."
            if formatName in text:
                text = text.replace(formatName, str(nameMap[name]))
                text = self._replace_names(nameMap, text)
        return text

    def _replace_params(self, nameMap, text):
        for name in nameMap:
            formatName = "&" + name + "."
            # print('_replace_params(): checking param = "' + formatName + '"')
            if formatName in text:
                # print('_replace_params(): found param = "' + formatName + '" in "' + text + '"')
                text = text.replace(formatName, str(nameMap[name]))
                text = self._replace_params(nameMap, text)
        return text.replace(' ', '')

    def _convert_param_names(self, param):
        if str(param).isdigit():
            return param
        else:
            var = self._replace_params(self._loopStack, param)
            var = self._replace_params(self.translationSteps['varDef'], var)
        return var
